/**
* @file application.c
* @brief example application.
* @author JK, ELV Elektronik AG
* @date 10.09.2021
* @version 1
**/
/*-----------------------------------------------------------------------------
* Copyright 2021 ELV Elektronik AG
* All Rights Reserved.
*
* The information contained in this file is confidential property of "ELV
* Elektronik AG". The use copying, transfer or disclosure is prohibited
* except by written agreement with "ELV Elektronik AG".
*----------------------------------------------------------------------------*/
#include "application.h"
#include "peripherals.h"

#include "lora_app.h"
#include "adc_if.h"			// SYS_GetSupplyLevel
#include "stm32_timer.h"

#include "stm32_seq.h"		// Sequenzer implementation to realize tasks
#include "utilities_def.h"	// Task ID to register and schedule application task


/*** private application functions ***/
void button_pressed( void );

void timer_wake_up( void* );
void enable_wake_up_timer( uint32_t sleeptime_ms );
void disable_wake_up_timer( void );
void reset_wake_up_timer( void );

void schedule_application( void );

void join( void );
void schedule_joining_procedure( void );



/**
 * @brief Timer to handle the button debounce
 */
static UTIL_TIMER_Object_t cycle_timer;

/**
 * @brief Used to send a msg if the button is not pushed for a defined time
 */
static uint32_t application_cycle = DEFAULT_APPLICATION_CYCLE;

/**
 * @brief Indicates an unsuccessful attempt to join the lorawan network
 */
static uint8_t joining_required = 0;

/**
 * @brief Counts unsuccessful joining attempts
 */
static uint8_t joining_attempts_cnt = 0;

/**
 * @brief Increasing delay times between unsuccessful joining attempts.
 * The last defined delay is still used if the number of unsuccessful joining attempts
 * exceeds the number of defined delays
 */
static uint16_t joining_retry_delay[] = {15, 30, 90};	// seconds

/**
 * @brief Number of increasing delay times between unsuccessful joining attempts
 */
static const uint8_t nb_joining_retry_delays = sizeof( joining_retry_delay ) / sizeof( joining_retry_delay[0] );


/**
 * @brief Initialize application related stuff.
 * Will be called in main.
 */
void application_init( void )
{
	UTIL_SEQ_RegTask( (1 << CFG_SEQ_Task_application), UTIL_SEQ_RFU, application_run );

	UTIL_TIMER_Create( &cycle_timer, 0xFFFFFFFFU, UTIL_TIMER_ONESHOT, timer_wake_up, NULL );

	button_init( button_pressed );
	leds_init();

	led_green_flash();

	register_joining_success_callback( application_run );
	register_joining_failed_callback( schedule_joining_procedure );

	register_tx_success_callback( led_green_flash );
	register_tx_failed_callback( led_red_flash );
}


/**
 * @brief Event or time based application implementation.
 * Schedule execution via schedule_application()
 */
void application_run( void )
{
	led_green_flash(); 		// signals application execution

	schedule_transmission( UNCONFIRMED_TRANSMISSION );

	enable_wake_up_timer( application_cycle );
}


/**
 * @brief Define application payload here. Will be called in SendTxData (lora_app.c)
 * @param buffer Pointer to array to store the payload
 * @return Size of payload
 */
uint8_t build_application_payload( uint8_t* buffer )
{
	uint8_t payload_size = 0;


	uint16_t bat = SYS_GetSupplyLevel();

	buffer[payload_size++] = (uint8_t) (bat >> 8);
	buffer[payload_size++] = (uint8_t) (bat);


	return payload_size;
}

/**
 * @brief Application downlink data
 * @param downlink_data Array that contains the received payload
 * @param downlink_size Number of received byte
 */
void application_downlink( uint8_t* downlink_data, uint8_t downlink_size )
{

}

/**
 * @brief User Button callback function
 */
void button_pressed( void )
{
	if( is_loramac_joined() )
	{
		schedule_transmission( UNCONFIRMED_TRANSMISSION );
	}

	reset_wake_up_timer();
}


/**
 * @brief Timer callback function
 * @param unused Argument needed to match required prototype
 */
void timer_wake_up( void* unused )
{
	if( !joining_required )
	{
		schedule_transmission( CONFIRMED_TRANSMISSION );

		enable_wake_up_timer( application_cycle );
	}
	else
	{
		join();
	}
}


/**
 * @brief Enables the wake up timer which calls the function timer_wake_up()
 * @param sleeptime_ms Time till wake up
 */
void enable_wake_up_timer( uint32_t sleeptime_ms )
{
	UTIL_TIMER_SetPeriod( &cycle_timer, sleeptime_ms );
	UTIL_TIMER_Start( &cycle_timer );
}


/**
 * @brief Disables the wake up timer
 */
void disable_wake_up_timer( void )
{
	UTIL_TIMER_Stop( &cycle_timer );
}


/**
 * @brief Restarts the wake up timer with the time stored in application_cycle
 */
void reset_wake_up_timer( void )
{
	disable_wake_up_timer();
	enable_wake_up_timer( application_cycle );
}


/**
 * @brief Schedules the application task
 */
void schedule_application( void )
{
  UTIL_SEQ_SetTask( ( 1 << CFG_SEQ_Task_application ), CFG_SEQ_Prio_0 );
}


/**
 * @brief Schedules another joining attempt
 */
void schedule_joining_procedure( void )
{
	led_red_flash(); // signals unsuccessful joining attempt

	uint8_t delay_nb = joining_attempts_cnt++;

	if( delay_nb >= nb_joining_retry_delays )
	{
		delay_nb = nb_joining_retry_delays - 1;
	}

	enable_wake_up_timer( SEC_2_MS(joining_retry_delay[delay_nb]) );

	joining_required = 1;
}


/**
 * @brief (Re)joining
 */
void join( void )
{
	joining_required = 0;

	led_green_flash();

	loramac_join();
}
