/**
* @file peripherals.c
* @brief implementations for example peripherals.
* @author JK, ELV Elektronik AG
* @date 10.09.2021
* @version 1
**/
/*-----------------------------------------------------------------------------
* Copyright 2021 ELV Elektronik AG
* All Rights Reserved.
*
* The information contained in this file is confidential property of "ELV
* Elektronik AG". The use, copying, transfer or disclosure is prohibited
* except by written agreement with "ELV Elektronik AG".
*----------------------------------------------------------------------------*/
#include "peripherals.h"
#include "hw_conf.h"
#include "hw_gpio.h"

#include "stm32_timer.h"
#include "stm32_seq.h"
#include "utilities_def.h"


/**
 * @brief Timer to handle the button debounce
 */
static UTIL_TIMER_Object_t debounce_timer;

/**
 * @brief Debouncing state, check for stable state
 */
static uint16_t bounce_state = 0;


void (*button_triggered_callback)( void ) = NULL;



void button_init( void (*button_triggered_cb)( void ) )
{
	button_triggered_callback = button_triggered_cb;

	UTIL_TIMER_Create(&debounce_timer, 0xFFFFFFFFU, UTIL_TIMER_ONESHOT, debounce_button, NULL);
	UTIL_TIMER_SetPeriod(&debounce_timer, BUTTON_DEBOUNCE_PERIOD);


	GPIO_InitTypeDef initStruct = { 0 };

	initStruct.Mode = GPIO_MODE_IT_FALLING;
	initStruct.Pull = GPIO_PULLUP;
	initStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;

	HW_GPIO_Init( BUTTON_GPIO_PORT, BUTTON_GPIO_PIN, &initStruct);
	HW_GPIO_SetIrq( BUTTON_GPIO_PORT, BUTTON_GPIO_PIN, 10, state_button_cb);
}


void button_deinit( void )
{
	HW_GPIO_Deinit( BUTTON_GPIO_PORT, BUTTON_GPIO_PIN, 0 );
}


void debounce_button( void *context )
{
	bounce_state = (bounce_state << 1) | HW_GPIO_Read( BUTTON_GPIO_PORT, BUTTON_GPIO_PIN) | 0xe000;
	if (bounce_state == 0xF000)          // Button is LOW
	{
		if( button_triggered_callback != NULL )
		{
			button_triggered_callback();
		}

		HW_GPIO_enable_irq( BUTTON_GPIO_PIN);
	}
	else if (bounce_state == 0xFFFF)     // Button is HIGH
	{
		HW_GPIO_enable_irq( BUTTON_GPIO_PIN );
	}
	else
	{
		UTIL_TIMER_Start(&debounce_timer);
	}
}


void state_button_cb( void *context )
{
	HW_GPIO_disable_irq( BUTTON_GPIO_PIN);

	/* debounce timer */
	bounce_state = 1;

	UTIL_TIMER_Start(&debounce_timer);
}


void leds_init( void )
{
	GPIO_InitTypeDef initStruct = { 0 };

	initStruct.Mode = GPIO_MODE_OUTPUT_PP;
	initStruct.Pull = GPIO_NOPULL;
	initStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;

	HW_GPIO_Init( LED_GREEN_GPIO_PORT, LED_GREEN_GPIO_PIN, &initStruct);
	HW_GPIO_Init( LED_RED_GPIO_PORT, LED_RED_GPIO_PIN, &initStruct);

	HW_GPIO_Write( LED_GREEN_GPIO_PORT, LED_GREEN_GPIO_PIN, GPIO_PIN_SET);
	HW_GPIO_Write( LED_RED_GPIO_PORT, LED_RED_GPIO_PIN, GPIO_PIN_SET);
}


void leds_deinit( void )
{
	HW_GPIO_Deinit( LED_GREEN_GPIO_PORT, LED_GREEN_GPIO_PIN, 0 );
	HW_GPIO_Deinit( LED_RED_GPIO_PORT, LED_RED_GPIO_PIN, 0 );
}


void led_red_en( uint8_t en )
{
	HW_GPIO_Write( LED_RED_GPIO_PORT, LED_RED_GPIO_PIN, !en );
}


void led_red_toggle( void )
{
	HW_GPIO_Toggle( LED_RED_GPIO_PORT, LED_RED_GPIO_PIN );
}


void led_red_flash( void )
{
	led_red_en(1);
	HAL_Delay(20);
	led_red_en(0);
}


void led_green_en( uint8_t en )
{
	HW_GPIO_Write( LED_GREEN_GPIO_PORT, LED_GREEN_GPIO_PIN, !en );
}


void led_green_toggle( void )
{
	HW_GPIO_Toggle( LED_GREEN_GPIO_PORT, LED_GREEN_GPIO_PIN );
}


void led_green_flash( void )
{
	led_green_en(1);
	HAL_Delay(20);
	led_green_en(0);
}

