/**
  ******************************************************************************
  * @file    adc_if.c
  * @author  MCD Application Team
  * @brief   Read status related to the chip (battery level, VREF, chip temperature)
  ******************************************************************************
  * @attention
  *
  * <h2><center>&copy; Copyright (c) 2021 STMicroelectronics.
  * All rights reserved.</center></h2>
  *
  * This software component is licensed by ST under Ultimate Liberty license
  * SLA0044, the "License"; You may not use this file except in compliance with
  * the License. You may obtain a copy of the License at:
  *                             www.st.com/SLA0044
  *
  ******************************************************************************
  */

/* Includes ------------------------------------------------------------------*/
#include "adc_if.h"
#include "sys_app.h"

/* External variables ---------------------------------------------------------*/
/**
  * @brief ADC handle
  */
extern ADC_HandleTypeDef hadc;

/* Private typedef -----------------------------------------------------------*/

/* Private define ------------------------------------------------------------*/
#define TEMPSENSOR_TYP_CAL1_V          ( ( int32_t)  760 )        /*!< Internal temperature sensor, parameter V30 (unit: mV). Refer to device datasheet for min/typ/max values. */
#define TEMPSENSOR_TYP_AVGSLOPE        ( ( int32_t) 2500 )        /*!< Internal temperature sensor, parameter Avg_Slope (unit: uV/DegCelsius). Refer to device datasheet for min/typ/max values. */

/* Private macro -------------------------------------------------------------*/
/* Private variables ---------------------------------------------------------*/
/**
  * @brief This function reads the ADC channel
  * @param channel channel number to read
  * @return adc measured level value
  */
static uint32_t ADC_ReadChannels( uint32_t channel );

/**
  * @brief This function reads ADC
  * @param channel channel number to read
  * @return value voltage level in linear scale
  */
static uint16_t ADC_Read( uint32_t channel );

/* Exported functions --------------------------------------------------------*/
void SYS_InitMeasurement( void )
{
  hadc.Instance = ADC;
}

void SYS_DeInitMeasurement( void )
{
}

int16_t SYS_GetTemperatureLevel( void )
{
  int16_t temperatureDegreeC = 0;
  uint32_t measuredLevel = 0;
  uint16_t batteryLevelmV = SYS_GetSupplyLevel();

  measuredLevel = ADC_ReadChannels( ADC_CHANNEL_TEMPSENSOR );

  /* convert ADC level to temperature */
  /* Device with temperature sensor not calibrated in production: use generic parameters */
  temperatureDegreeC = __LL_ADC_CALC_TEMPERATURE_TYP_PARAMS(  TEMPSENSOR_TYP_AVGSLOPE,
                                                              TEMPSENSOR_TYP_CAL1_V,
                                                              TEMPSENSOR_CAL1_TEMP,
                                                              batteryLevelmV,
                                                              measuredLevel,
                                                              LL_ADC_RESOLUTION_12B );

  APP_LOG( TS_ON, VLEVEL_L, "temp= %d\n\r", temperatureDegreeC );

  /* from int16 to q8.7*/
  temperatureDegreeC <<= 8;

  return ( int16_t ) temperatureDegreeC;
}

uint16_t SYS_GetSupplyLevel( void )
{
  return ADC_Read( ADC_CHANNEL_VREFINT );
}

uint16_t SYS_GetBatteryLevel( uint32_t channel )
{
  return ADC_Read( channel );
}

/* Private Functions Definition -----------------------------------------------*/
static uint32_t ADC_ReadChannels( uint32_t channel )
{
  uint32_t ADCxConvertedValues = 0;
  ADC_ChannelConfTypeDef sConfig = { 0 };

  MX_ADC_Init();

  /* Start Calibration */
  if(HAL_ADCEx_Calibration_Start( &hadc ) != HAL_OK )
  {
    Error_Handler();
  }

  /* Configure Regular Channel */
  sConfig.Channel = channel;
  sConfig.Rank = ADC_REGULAR_RANK_1;
  sConfig.SamplingTime = ADC_SAMPLINGTIME_COMMON_1;
  if( HAL_ADC_ConfigChannel( &hadc, &sConfig ) != HAL_OK )
  {
    Error_Handler();
  }

  if( HAL_ADC_Start( &hadc ) != HAL_OK )
  {
    /* Start Error */
    Error_Handler();
  }
  /** Wait for end of conversion */
  HAL_ADC_PollForConversion( &hadc, HAL_MAX_DELAY );

  /** Wait for end of conversion */
  HAL_ADC_Stop( &hadc ) ;   /* it calls also ADC_Disable() */

  ADCxConvertedValues = HAL_ADC_GetValue( &hadc );

  HAL_ADC_DeInit( &hadc );

  return ADCxConvertedValues;
}

static uint16_t ADC_Read( uint32_t channel )
{
  uint8_t i, avg_cnt = 0, min_idx = 0, max_idx = 0;
  uint16_t battery_level_mv = 0;
  uint32_t measured_level[12] = { 0 }, measured_level_avg = 0;

  measured_level[0] = ADC_ReadChannels( channel );
  for( i = 1; i < 12; i++ )
  {
    measured_level[i] = ADC_ReadChannels( channel );
    if( measured_level[i] < measured_level[min_idx] )
    {
      min_idx = i;
    }
    else if( measured_level[i] > measured_level[max_idx] )
    {
      max_idx = i;
    }
  }
  for( i = 0; i < 12; i++ )
  {
    if( i != min_idx && i != max_idx )
    {
      measured_level_avg += measured_level[i];
      avg_cnt++;
    }
  }

  measured_level_avg /= avg_cnt;

  if( measured_level_avg == 0 )
  {
    battery_level_mv = 0;
  }
  else
  {
    /* Device with Reference voltage not calibrated in production: use generic parameters */
    if( channel == ADC_CHANNEL_VREFINT )
    {
      if( ( uint32_t )*VREFINT_CAL_ADDR != ( uint32_t )0xFFFFU )
      {
        battery_level_mv = __LL_ADC_CALC_VREFANALOG_VOLTAGE( measured_level_avg, ADC_RESOLUTION_12B );
      }
      else
      {
        battery_level_mv = ( VREFINT_CAL_VREF * 1510 ) / measured_level_avg;
      }
    }
    else
    {
      if( ( uint32_t )*VREFINT_CAL_ADDR != ( uint32_t )0xFFFFU )
      {
        battery_level_mv = __LL_ADC_CALC_DATA_TO_VOLTAGE( SYS_GetSupplyLevel(), measured_level_avg, ADC_RESOLUTION_12B );
      }
      else
      {
        battery_level_mv = ( VREFINT_CAL_VREF * measured_level_avg ) / 4095;  // If channel != VREFINT
      }
    }
  }

  return battery_level_mv;
}

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
