/**
 ******************************************************************************
 * @file    lora_app.c
 * @author  MCD Application Team
 * @brief   Application of the LRWAN Middleware
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; Copyright (c) 2021 STMicroelectronics.
 * All rights reserved.</center></h2>
 *
 * This software component is licensed by ST under Ultimate Liberty license
 * SLA0044, the "License"; You may not use this file except in compliance with
 * the License. You may obtain a copy of the License at:
 *                             www.st.com/SLA0044
 *
 ******************************************************************************
 */

/* Includes -------------------------------------------------------------------*/
#include "main.h"
#include "Region.h" /* Needed for LORAWAN_DEFAULT_DATA_RATE */
#include "sys_app.h"
#include "lora_app.h"
#include "stm32_seq.h"
#include "stm32_timer.h"
#include "utilities_def.h"
#include "lora_app_version.h"
#include "lorawan_version.h"
#include "subghz_phy_version.h"
#include "lora_info.h"
#include "LmHandler.h"
#include "stm32_lpm.h"
#include "adc_if.h"
#include "sys_conf.h"


/* USER CODE BEGIN Includes */
#include "application.h"	// get_application_payload

/* USER CODE END Includes */

/* External variables ---------------------------------------------------------*/
/* Private typedef ------------------------------------------------------------*/
/* Private define -------------------------------------------------------------*/

/* Private macro --------------------------------------------------------------*/
/* Private function prototypes ------------------------------------------------*/
/**
 * @brief  LoRa endNode send request
 * @param  none
 * @retval none
 */
static void SendTxData(void);

/**
 * @brief  join event callback function
 * @param  params
 * @retval none
 */
static void OnJoinRequest(LmHandlerJoinParams_t *joinParams);

/**
 * @brief  join success callback function
 * @param  params
 * @retval none
 */
static void (*OnJoinSuccess)( void ) = NULL;

/**
 * @brief  join failed callback function
 * @param  params
 * @retval none
 */
static void (*OnJoinFailed)( void ) = NULL;

/**
 * @brief  tx success callback function. Will be triggered if LoRaMAC has sent the msg
 * @param  params
 * @retval none
 */
static void (*OnTxSuccess)( void ) = NULL;

/**
 * @brief  tx failed callback function. Will be triggered if LoRaMAC is not ready to send
 * @param  params
 * @retval none
 */
static void (*OnTxFailed)( void ) = NULL;

/**
 * @brief  tx event callback function
 * @param  params
 * @retval none
 */
static void OnTxData(LmHandlerTxParams_t *params);

/**
 * @brief callback when LoRa endNode has received a frame
 * @param appData
 * @param params
 * @retval None
 */
static void OnRxData(LmHandlerAppData_t *appData, LmHandlerRxParams_t *params);

/*!
 * Will be called each time a Radio IRQ is handled by the MAC layer
 *
 */
static void OnMacProcessNotify(void);


/* Private variables ---------------------------------------------------------*/
/**
 * @brief Next TX slot in nextTxIn ms.
 */
UTIL_TIMER_Time_t nextTxIn = 0;

/**
 * @brief Timestamp, recognized LOW (Button)
 */
SysTime_t low_timestamp = { .Seconds = 0, .SubSeconds = 0 };

/**
 * @brief Timestamp, recognized HIGH (Button)
 */
SysTime_t high_timestamp = { .Seconds = 0, .SubSeconds = 0 };

/**
 * @brief User application buffer
 */
static uint8_t AppDataBuffer[LORAWAN_APP_DATA_BUFFER_MAX_SIZE];

/**
 * @brief User application data structure
 */
static LmHandlerAppData_t AppData = { 0, 0, AppDataBuffer };

static ActivationType_t ActivationType = LORAWAN_DEFAULT_ACTIVATION_TYPE;

/**
 * @brief LoRaWAN handler Callbacks
 */
static LmHandlerCallbacks_t LmHandlerCallbacks =
{
		.GetBatteryLevel = GetBatteryLevel, .GetTemperature = GetTemperatureLevel,
		.OnMacProcess = OnMacProcessNotify, .OnJoinRequest = OnJoinRequest,
		.OnTxData = OnTxData, .OnRxData = OnRxData
};

/**
 * @brief LoRaWAN handler parameters
 */
static LmHandlerParams_t LmHandlerParams =
{
		.ActiveRegion = ACTIVE_REGION, .DefaultClass = LORAWAN_DEFAULT_CLASS,
		.AdrEnable = LORAWAN_ADR_STATE, .TxDatarate = LORAWAN_DEFAULT_DATA_RATE,
		.PingPeriodicity = LORAWAN_DEFAULT_PING_SLOT_PERIODICITY
};

/**
  * @brief LoRaWAN message type
  */
static LmHandlerMsgTypes_t lora_msg_type = LORAWAN_DEFAULT_CONFIRMED_MSG_STATE;



void LoRaWAN_Init(void)
{
	APP_LOG(TS_OFF, VLEVEL_M, "LoRaWAN_Init\r\n");

	UTIL_SEQ_RegTask((1 << CFG_SEQ_Task_LmHandlerProcess), UTIL_SEQ_RFU, LmHandlerProcess);
	UTIL_SEQ_RegTask((1 << CFG_SEQ_Task_LoRaSendOnTxTimerOrButtonEvent), UTIL_SEQ_RFU, SendTxData);

	/* Init Info table used by LmHandler*/
	LoraInfo_Init();

	/* Init the Lora Stack*/
	LmHandlerInit(&LmHandlerCallbacks);

	LmHandlerConfigure(&LmHandlerParams);

	LmHandlerJoin(ActivationType);
}

/* Private functions ---------------------------------------------------------*/

static void OnRxData(LmHandlerAppData_t *appData, LmHandlerRxParams_t *params)
{
	// OnRxData will be called after OnTxData
	if ((appData != NULL) && (params != NULL))
	{
#if defined (APP_LOG_ENABLED) && (APP_LOG_ENABLED == 1)
		static const char *slotStrings[] =
		{ "1", "2", "C", "C Multicast", "B Ping-Slot", "B Multicast Ping-Slot" };
#endif
		APP_LOG(TS_ON, VLEVEL_M, "\r\n###### ========== MCPS-Indication ==========\r\n");
		APP_LOG(TS_ON, VLEVEL_M,
				"###### D/L FRAME:%04d | SLOT:%s | PORT:%d | DR:%d | RSSI:%d | SNR:%d\r\n",
				params->DownlinkCounter, slotStrings[params->RxSlot],
				appData->Port, params->Datarate, params->Rssi, params->Snr);

		switch (appData->Port) {
		case LORAWAN_USER_APP_PORT:
			application_downlink(appData->Buffer, appData->BufferSize);
			APP_LOG(TS_OFF, VLEVEL_H, "Downlink\r\n");
			break;
		default:
			break;
		}
	}
}

static void SendTxData(void)
{
	APP_LOG( TS_ON, VLEVEL_L, "SendTxData\r\n" );

	AppData.Port = LORAWAN_USER_APP_PORT;

	AppData.BufferSize = build_application_payload(AppData.Buffer);

	nextTxIn = 0;

	if ( LORAMAC_HANDLER_SUCCESS == LmHandlerSend( &AppData, LORAWAN_DEFAULT_CONFIRMED_MSG_STATE, &nextTxIn, false ) )
	{
		APP_LOG( TS_ON, VLEVEL_L, "SEND REQUEST\r\n" );
		if( OnTxSuccess != NULL )
		{
			OnTxSuccess();
		}
	}
	else
	{
		if( OnTxFailed != NULL )
		{
			OnTxFailed();
		}
	}

}


static void OnTxData(LmHandlerTxParams_t *params)
{
	if ((params != NULL) && (params->IsMcpsConfirm != 0))
	{
		APP_LOG(TS_OFF, VLEVEL_M, "\r\n###### ========== MCPS-Confirm =============\r\n");
		APP_LOG(TS_OFF, VLEVEL_H,
				"###### U/L FRAME:%04d | PORT:%d | DR:%d | PWR:%d",
				params->UplinkCounter, params->AppData.Port, params->Datarate,
				params->TxPower);

		APP_LOG(TS_OFF, VLEVEL_H, " | MSG TYPE:");
		if (params->MsgType == LORAMAC_HANDLER_CONFIRMED_MSG)
		{
			if (params->AckReceived != 0)
			{
				// ...
			}
			else
			{
				// ...
			}
			APP_LOG(TS_OFF, VLEVEL_H, "CONFIRMED [%s]\r\n", (params->AckReceived != 0) ? "ACK" : "NACK");
		}
		else
		{
			APP_LOG(TS_OFF, VLEVEL_H, "UNCONFIRMED\r\n");
		}
	}
}

static void OnJoinRequest(LmHandlerJoinParams_t *joinParams)
{
	if (joinParams != NULL)
	{
		if (joinParams->Status == LORAMAC_HANDLER_SUCCESS)
		{
			APP_LOG(TS_OFF, VLEVEL_M, "\r\n###### = JOINED = ");
			if (joinParams->Mode == ACTIVATION_TYPE_ABP)
			{
				APP_LOG(TS_OFF, VLEVEL_M, "ABP ======================\r\n");
			}
			else
			{
				APP_LOG(TS_OFF, VLEVEL_M, "OTAA =====================\r\n");
			}

			if( OnJoinSuccess != NULL )
			{
				OnJoinSuccess();
			}

		}
		else
		{
			APP_LOG(TS_OFF, VLEVEL_M, "\r\n###### = JOIN FAILED\r\n");

			if( OnJoinFailed != NULL )
			{
				OnJoinFailed();
			}
		}
	}

	/* send every time timer elapses */
}

static void OnMacProcessNotify(void)
{
	UTIL_SEQ_SetTask((1 << CFG_SEQ_Task_LmHandlerProcess), CFG_SEQ_Prio_0);
}




/************************ (C) COPYRIGHT STMicroelectronics *********/

void loramac_join( void )
{
	LmHandlerJoin(ActivationType);
}

uint8_t is_loramac_busy( void )
{
	return LmHandlerIsBusy();
}

uint8_t is_loramac_joined( void )
{
	return (LmHandlerJoinStatus() == LORAMAC_HANDLER_SET);
}

void register_joining_success_callback( void (*on_joining_success)( void ) )
{
	OnJoinSuccess = on_joining_success;
}

void register_joining_failed_callback( void (*on_joining_failed)( void ) )
{
	OnJoinFailed = on_joining_failed;
}

void register_tx_success_callback( void (*on_joining_success)( void ) )
{
	OnTxSuccess = on_joining_success;
}

void register_tx_failed_callback( void (*on_joining_failed)( void ) )
{
	OnTxFailed = on_joining_failed;
}

void schedule_transmission( uint8_t confirmed )
{
  if( confirmed )
  {
    lora_msg_type = LORAMAC_HANDLER_CONFIRMED_MSG;
  }
  else
  {
    lora_msg_type = LORAMAC_HANDLER_UNCONFIRMED_MSG;
  }

  UTIL_SEQ_SetTask( ( 1 << CFG_SEQ_Task_LoRaSendOnTxTimerOrButtonEvent ), CFG_SEQ_Prio_0 );
}


